# -*- coding: utf-8 -*-

"""Package with CSV renderer and parser for SlashDB."""
import csv

import logging
import io

from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPNotImplemented,
    HTTPCreated,
)
from pyramid.response import Response
from slashdb.exceptions import NoParams, UnexpectedParameters, MissingRequiredParameters, MissingValueForParameter

from slashdb.formats import (
    DefaultIsStreamingMixin, IParser, IRenderer,
    check_for_ambiguous_keys, get_href_param,
)
from slashdb.formats.csv_sdb.decoder import CSVDecoder, UTF8Recorder
from slashdb.formats.csv_sdb.emitter import CSVEmitter
from slashdb.formats.csv_sdb.encoder import CSVEncoder
from slashdb.querydef_utils import make_query_url_template
from slashdb.utils import to_unicode

__version__ = '0.0.1'

log = logging.getLogger('slashdb')

EXTENSIONS = ('csv', )
MIME_TYPES = ('text/csv',)


truthy = frozenset(('t', 'true', 'y', 'yes', 'on', '1'))
falsy = frozenset(('f', 'false', 'n', 'no', 'off', '0'))


class CSVRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    default_extension = EXTENSIONS[0]
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'csv_sdb.stream_by_default'
    href_enabled_default = True
    href_ini_param_str = 'csv_sdb.href'
    headers_enabled_default = 'true'
    headers_ini_param_str = 'csv_sdb.headers_by_default'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        unicode_explanation = to_unicode(exc.explanation)
        data = {
            'http_code': exc.status_code,
            'description': unicode_explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            data['url_template'] = make_query_url_template(query_config, 'csv')
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        data.update(kwargs)

        sio = io.StringIO()
        writer = csv.DictWriter(sio, fieldnames=list(data.keys()))
        writer.writeheader()
        writer.writerow(data)
        body = sio.getvalue()

        explanation = cls._encode_explanation(unicode_explanation)
        response = Response(
            body=body,
            status_int=exc.code,
            content_type=cls.content_type,
            charset=output_encoding,
        )
        response.headers.add('Warning', explanation)

        return response

    @classmethod
    def content_location_response(cls, locations, request):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        if locations:
            locations = [f'{l}.{cls.default_extension}' for l in locations]
            data = [{'Content-Location': loc} for loc in locations]

            sio = io.StringIO()
            writer = csv.DictWriter(sio, fieldnames=list(data[0].keys()))
            writer.writeheader()
            for d in data:
                writer.writerow(d)

            body = sio.getvalue()
            response = HTTPCreated(
                body=body,
                content_location=locations[0] if len(locations) == 1 else None,
                content_type=cls.content_type,
                charset=output_encoding
            )
        else:
            response = HTTPCreated(
                charset=output_encoding,
                content_type=cls.content_type,
            )

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_encoding = context.db_config.db_encoding
        self.res_req_meta = data_proxy.res_req_meta
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.ini_settings = request.registry.settings
        self.sdb_settings = self.ini_settings['sdb_settings']
        self.output_encoding = self.sdb_settings['output_encoding']

        self._enable_href = get_href_param(
            request.params,
            self.ini_settings.get(
                self.href_ini_param_str,
                self.href_enabled_default
            )
        )

    @property
    def with_headers(self):
        """Checks if CSV output should contain headers in first line.

         First url param then renderer settings.

        """
        url_param = self.request.params.get('headers')
        renderer_default = self.ini_settings.get(
            self.headers_ini_param_str,
            self.headers_enabled_default
        )

        def catch_bool_parse_exception(v):
            """Wrapper around self.asbool."""
            try:
                return self.asbool(v)
            except ValueError:
                raise HTTPBadRequest(explanation='Headers has to be True or False.')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter headers is missing value (true or false).")
        elif url_param:
            return catch_bool_parse_exception(url_param)
        else:
            return catch_bool_parse_exception(renderer_default)

    @staticmethod
    def asbool(s):
        """ Return the boolean value ``True`` if the case-lowered value of string
        input ``s`` is any of ``t``, ``true``, ``y``, ``on``, or ``1``, otherwise
        return the boolean value ``False``.  If ``s`` is the value ``None``,
        return ``False``.  If ``s`` is already one of the boolean values ``True``
        or ``False``, return it."""
        if s is None:
            return False
        if isinstance(s, bool):
            return s
        s = str(s).strip().lower()
        if s in truthy:
            return True
        elif s in falsy:
            return False
        raise ValueError()

    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        if self.is_streaming_response():
            if self.depth > 0:
                raise HTTPNotImplemented(
                    explanation="Streamed rendering to CSV of nested resources is not implemented."
                )

            return CSVEmitter(self.data_proxy, self.with_headers, self.db_encoding, self.output_encoding)
        else:
            csv_encoder = CSVEncoder(
                self.data_proxy, self.with_headers, self.db_encoding, self.output_encoding, flatten=self.depth > 0
            )
            return csv_encoder.encode()


class CSVParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.db_name = context.db_id
        self.last_res_req_meta = data_proxy.res_req_meta[-1]
        self.csv_null_str = request.params.get(
            'csvNullStr', request.registry.settings.get('csv_sdb.csv_null_str', None)
        )
        self.input_encoding = request.charset

    def parse(self, data_str):
        csv_decoder = CSVDecoder(self.db_name, self.last_res_req_meta, self.csv_null_str, self.input_encoding)
        update_dict = csv_decoder.decode(data_str)
        return update_dict

    @staticmethod
    def parse_sqlpassthru_parameters(required_params, request_body, request_charset, request_get_params, ini_settings):

        csv_null_str = request_get_params.get('csvNullStr', ini_settings.get('csv_sdb.csv_null_str', None))

        datafile = io.StringIO(request_body)
        csv_reader = csv.reader(datafile, 'excel')

        try:
            headers = next(csv_reader)
        except StopIteration:
            if required_params:
                raise NoParams("No data to parse found in incoming CSV.")
            else:
                return []

        if required_params:
            unexpected_params = set(headers) - set(required_params)
            if unexpected_params:
                raise UnexpectedParameters('Unexpected SQL Pass-thru parameters in incoming CSV.', unexpected_params)

            missing_params = set(required_params) - set(headers)
            if missing_params:
                raise MissingRequiredParameters('Missing required parameters in incoming CSV.', missing_params)

            body_params = []

            for line in csv_reader:
                if len(headers) != len(line):
                    raise MissingValueForParameter("Mismatch between number of headers and columns in incoming CSV.")

                # create single dict with parameters, replace null string if required
                param_set = {
                    key: None if csv_null_str and value == csv_null_str else value
                    for key, value in zip(headers, line)
                }
                body_params.append(param_set)

            if body_params:
                return body_params
            elif headers:
                raise MissingRequiredParameters(
                    "Missing SQL Pass-thru parameters values, only header found in incoming CSV.", headers
                )
        elif headers:
            raise UnexpectedParameters(
                "Unexpected data found in incoming CSV for SQLPass-thru that doesn't require parameters.", headers
            )
        else:
            return []


Renderer = CSVRenderer
Parser = CSVParser
