# -*- coding: utf-8 -*-

"""Package with HTML renderer for SlashDB."""

from pyramid.renderers import render, render_to_response
from pyramid.httpexceptions import HTTPNotImplemented

from slashdb.querydef_utils import make_query_url_template
from slashdb.formats import DefaultIsStreamingMixin, IRenderer, check_for_ambiguous_keys
from slashdb.formats.html_sdb.encoder import HTMLEncoder
from slashdb.utils import to_unicode

__version__ = '0.0.1'

EXTENSIONS = ('html',)
MIME_TYPES = ('text/html',)


class HTMLRenderer(DefaultIsStreamingMixin, IRenderer):
    """HTML renderer for SlashDB.
    Class of SlashDB IRenderer that gives a standard user interface.

    Renderer is configurable in .ini settings with parameters:

    mako.directories - location of mako templates, by default slashdb.formats.html_sdb:templates
    static_path - location of static files like css, js, images etc., by default slashdb.formats.html_sdb:static

    """

    extensions = EXTENSIONS
    default_extension = EXTENSIONS[0]
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'html_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        sdb_settings = request.registry.settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        unicode_explanation = to_unicode(exc.explanation)
        data = {
            'code': exc.code,
            'title': exc.title,
            'detail': exc.detail,
            'explanation': unicode_explanation,
            'query_url_template': ''
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            data['query_url_template'] = make_query_url_template(query_config, 'html')
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        data.update(kwargs)

        response = render_to_response('/http_exception.mako', data, request=request)
        explanation = cls._encode_explanation(unicode_explanation)
        response.headers['Warning'] = explanation
        response.status_code = exc.status_code
        response.content_type = cls.content_type
        response.charset = output_encoding
        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_id = context.db_id
        self.query_id = context.query_id
        self.url_segments = context.url_segments
        self.db_encoding = context.db_config.db_encoding
        self.build_escaped_url = context.get_replace_fun()

        self.res_req_meta = data_proxy.res_req_meta

        self.url_params = request.params
        self.ini_settings = request.registry.settings
        sdb_settings = request.registry.settings['sdb_settings']
        self.output_encoding = sdb_settings['output_encoding']

    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        if self.is_streaming_response():
            raise HTTPNotImplemented(explanation="Streaming HTML is not implemented.")

        # task #159 transpose ignored in html
        self.data_proxy.transpose = False

        data = self.data_proxy.get(stream=False)
        html_encoder = HTMLEncoder(
            self.url_segments,
            self.url_params,
            self.data_proxy,
            self.db_id,
            self.query_id,
            self.db_encoding,
            self.output_encoding,
            self.build_escaped_url,
            self.context.reversed_url_substitution,
        )
        template, context = html_encoder.encode(data)
        return render(template, context, request=self.request).encode(self.output_encoding)

    def response_postprocessing(self, response):
        pass


Renderer = HTMLRenderer
Parser = None
