# -*- coding: utf-8 -*-

"""Package with JSON renderer and parser for SlashDB."""
from itertools import chain
from typing import List

from pyramid.httpexceptions import HTTPCreated
from pyramid.response import Response
from pyramid.settings import asbool

from slashdb.exceptions import NoParams, UnexpectedParameters, MissingRequiredParameters

from slashdb.formats import (
    DefaultIsStreamingMixin,
    IParser,
    IRenderer,
    check_for_ambiguous_keys,
    get_href_param,
    is_numeric
)
from slashdb.formats.json_sdb.decoder import JSONDecoder, simplejson
from slashdb.formats.json_sdb.emitter import JSONEmitter
from slashdb.querydef_utils import make_query_url_template
from slashdb.utils import to_unicode
from slashdb.utils import get_message

__version__ = '0.0.1'

EXTENSIONS = ('json', )
MIME_TYPES = ('application/json', )
# custom charset not allowed for application/json
JSON_ENCODING = 'utf-8'


class JSONRenderer(DefaultIsStreamingMixin, IRenderer):
    """JSON renderer for SlashDB."""

    extensions = EXTENSIONS
    default_extension = EXTENSIONS[0]
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'json_sdb.stream_by_default'
    href_enabled_default = True
    href_ini_param_str = 'json_sdb.href'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        sdb_settings = request.registry.settings['sdb_settings']
        output_encoding = JSON_ENCODING
        unicode_explanation = to_unicode(exc.explanation)

        data = {
            'http_code': exc.status_code,
            'description': unicode_explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            data['url_template'] = make_query_url_template(query_config, 'json')
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        data.update(kwargs)

        body = simplejson.dumps(data)
        explanation = cls._encode_explanation(unicode_explanation)
        response = Response(
            body=body,
            status_int=exc.code,
            content_type=cls.content_type,
            charset=output_encoding
        )
        response.headers.add('Warning', explanation)

        return response

    @classmethod
    def content_location_response(cls, locations: List[str], request):
        if locations:
            locations = [f'{l}.{cls.default_extension}' for l in locations]
            body = [{'Content-Location': loc} for loc in locations]
            body = simplejson.dumps(body)
            response = HTTPCreated(
                body=body,
                content_location=locations[0] if len(locations) == 1 else None,
                content_type=cls.content_type,
                charset=JSON_ENCODING
            )
        else:
            response = HTTPCreated(
                charset=JSON_ENCODING,
                content_type = cls.content_type,
            )

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_name = context.db_id
        self.db_config = context.db_config
        self.db_encoding = context.db_config.db_encoding
        self.query_id = context.query_id
        self.build_escaped_url = context.get_replace_fun()

        self.res_req_meta = data_proxy.res_req_meta

        self.host_url = request.host_url
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.ini_settings = request.registry.settings
        self.output_encoding = JSON_ENCODING

        self.indent = self.ini_settings.get('json_sdb.indent', None)
        self.minimize = asbool(
            self.ini_settings.get('json_sdb.minimize', False)
        )
        if self.minimize:
            self.indent = None

        if self.indent is not None:
            self.indent = ' ' * int(self.indent)

        self.emtr = None

        self._enable_href = get_href_param(
            request.params,
            self.ini_settings.get(
                self.href_ini_param_str,
                self.href_enabled_default
            )
        )

    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        is_streaming = self.is_streaming_response()
        self.emtr = JSONEmitter(
            self.data_proxy,
            self.db_encoding,
            self.indent,
            self.minimize,
            self.depth,
            self.build_escaped_url,
            is_streaming,
            self._enable_href,
        )

        if is_streaming:
            return self.emtr

        return self.emtr._to_string()

    def response_postprocessing(self, response):
        request_params = set(list(self.request.headers.keys()) + list(self.request.GET.keys()))

        if 'columns' in request_params:
            column_names = list(self.emtr.first_row.keys()) if self.emtr.first_row else []
            cn = simplejson.dumps(column_names)
            response.headers.add('columns', cn)

        if 'numeric_columns' in request_params:
            row = self.emtr.first_row or {}
            if 'transpose' in request_params:
                numeric_columns = {}
                for key, value_list in list(row.items()):
                    try:
                        numeric_columns[key] = is_numeric(value_list[0])
                    except IndexError:
                        numeric_columns[key] = False
            else:
                numeric_columns = {key: is_numeric(value) for key, value in row.items()}

            header_value = simplejson.dumps(numeric_columns)
            response.headers.add('numeric_columns', header_value)


class JSONParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.res_req_meta = data_proxy.res_req_meta
        self.input_encoding = request.charset

    def parse(self, data_str):
        json_decoder = JSONDecoder(self.res_req_meta, self.input_encoding)
        update_dict = json_decoder.decode(data_str)
        return update_dict

    @staticmethod
    def parse_sqlpassthru_parameters(required_params, request_body, request_charset, request_get_params, ini_settings):
        required_params = set(required_params)

        if not request_body.strip():
            if required_params:
                raise NoParams("No data to parse found in incoming JSON.")
            else:
                return []

        try:
            parsed_data = simplejson.loads(request_body, request_charset)
        except simplejson.JSONDecodeError as json_exc:
            raise NoParams(get_message(json_exc))

        # allow passing single param set
        if isinstance(parsed_data, dict):
            parsed_data = [parsed_data]

        try:
            found_param_names = set(chain(*(list(param_set.keys()) for param_set in parsed_data)))
        except AttributeError:
            raise NoParams(
                "Unexpected data structure in incoming JSON. "
                "Expected is a list of dictionaries with parameters for SQL Pass-thru.",
            )

        if required_params:
            if isinstance(parsed_data, list):
                unexpected_params = set(found_param_names) - set(required_params)
                if unexpected_params:
                    raise UnexpectedParameters(
                        'Unexpected SQL Pass-thru parameters in incoming JSON.',
                        unexpected_params
                    )

                missing_params = set(required_params) - set(found_param_names)
                if missing_params:
                    raise MissingRequiredParameters(
                        "Missing SQL Pass-thru parameters in incoming JSON.",
                        missing_params
                    )

                return parsed_data
            else:
                raise NoParams(
                    "Unexpected data structure in incoming JSON. "
                    "Expected is a list of dictionaries with parameters for SQL Pass-thru."
                )

        elif parsed_data:
            raise UnexpectedParameters(
                "Unexpected data found in incoming JSON for SQLPass-thru that doesn't require parameters.",
                found_param_names
            )
        else:
            return []


Renderer = JSONRenderer
Parser = JSONParser
