# -*- coding: utf-8 -*-

"""Package with TXT renderer and parser for SlashDB."""

from pyramid.settings import asbool
from pyramid.response import Response
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotImplemented

from slashdb.formats import DefaultIsStreamingMixin, IRenderer, IParser

from slashdb.formats.txt_sdb.encoder import TXTEncoder
from slashdb.formats.txt_sdb.decoder import TXTDecoder
from slashdb.querydef_utils import make_query_url_template
from slashdb.utils import to_unicode

__version__ = '0.0.1'


class TXTRenderer(DefaultIsStreamingMixin, IRenderer):
    """JSON renderer for SlashDB."""

    extensions = ('txt', )
    mime_types = ('text/plain', )
    renderable = ('value_single', 'value_many')
    content_type = mime_types[0]
    stream_by_default_str = 'txt_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        sdb_settings = request.registry.settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']

        unicode_explanation = to_unicode(exc.explanation)
        body_args = [
            'http_code: {}'.format(exc.status_code),
            'description: {}'.format(unicode_explanation)
        ]

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            body_args.append('url_template: {}'.format(make_query_url_template(query_config, 'txt')))
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        for key, value in list(kwargs.items()):
            body_args.append('{}: {}'.format(key, value))

        body = ', '.join(body_args)

        explanation = cls._encode_explanation(unicode_explanation)
        response = Response(
            body=body,
            status_int=exc.code,
            content_type=cls.content_type,
            charset=output_encoding,
        )
        response.headers.add('Warning', explanation)
        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_encoding = context.db_config.db_encoding
        self.db_name = context.db_id
        self.res_req_meta = data_proxy.res_req_meta
        self.host_url = request.host_url
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.ini_settings = request.registry.settings
        sdb_settings = request.registry.settings['sdb_settings']
        self.output_encoding = sdb_settings['output_encoding']

    @property
    def nil_visible(self):
        url_param = self.request.params.get('nil_visible')
        renderer_default = self.ini_settings.get('txt_sdb.nil_visible')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter nil_visible is missing value (true or false).")
        elif url_param:
            return asbool(url_param)
        elif renderer_default:
            return asbool(renderer_default)
        else:
            return False

    def render(self):
        if self.is_streaming_response():
            raise HTTPNotImplemented(explanation="Streaming TXT is not implemented.")

        # task #159 transpose ignored in html
        self.data_proxy.transpose = False

        resources = self.data_proxy.get(stream=False)
        body = TXTEncoder(self.output_encoding, self.db_encoding, self.nil_visible).encode(resources)
        return body

    def response_postprocessing(self, response):
        pass


class TXTParser(IParser):
    """Parser of text into more useful python value.

    As class to keep Maker - Parser logic.

    """

    extensions = ('txt', )
    mime_types = ('text/plain', )

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        res_req_meta = data_proxy.res_req_meta[-1]
        self.resource_class = res_req_meta.resource_class
        self.scalars = res_req_meta.scalars
        self.input_encoding = request.charset

    def parse(self, data_str):
        return TXTDecoder(self.resource_class, self.scalars, self.input_encoding).decode(data_str)


Renderer = TXTRenderer
Parser = TXTParser
