# -*- coding: utf-8 -*-

"""XML renderer and parser for SlashDB.

Module provides class of SlashDB IRenderer capable of returning pyramid response with SlashDB resources encoded in XML.


Renderer is configurable in settings with parameters:

xml_sdb.root_tag - sets root tag in xml, by default: SlashDB
xml_sdb.xmlns - namespace for XML instance, by default: http://www.vtenterprise.com/slashdb

"""
from itertools import chain

import logging
import xml.etree.ElementTree as ET

from pyramid.httpexceptions import HTTPBadRequest, HTTPCreated
from pyramid.response import Response
from pyramid.settings import asbool

from slashdb.exceptions import NoParams, UnexpectedParameters, MissingRequiredParameters

from slashdb.formats import (
    DefaultIsStreamingMixin, IParser, IRenderer,
    check_for_ambiguous_keys, get_href_param,
)
from slashdb.formats.xml_sdb.decoder import XMLDecoder
from slashdb.formats.xml_sdb.emitter import XMLEmitter, XMLRecordsetEmitter
from slashdb.formats.xsd_sdb import XSDEncoder
from slashdb.querydef_utils import make_query_url_template
from slashdb.utils import to_unicode, get_message

__version__ = '0.0.1'

log = logging.getLogger('slashdb')

EXTENSIONS = ('xml', )
MIME_TYPES = ('text/xml', 'application/xml')


class XMLRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    default_extension = EXTENSIONS[0]
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'xml_sdb.stream_by_default'
    href_enabled_default = True
    href_ini_param_str = 'xml_sdb.href'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']

        unicode_explanation = to_unicode(exc.explanation)
        root_tag = ini_settings.get('xml_sdb.root_tag', 'SlashDB'),
        tags = [
            ('http_code', exc.status_code),
            ('description', unicode_explanation)
        ]

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            tags.append(
                ('url_template', make_query_url_template(query_config, 'xml'))
            )
        except (AttributeError, TypeError, KeyError):
            pass  # then skip additional data

        for key_value in list(kwargs.items()):
            tags.append(key_value)

        body = [
            '<?xml version="1.0" encoding="{}"?>'.format(output_encoding),
            '<%s>' % root_tag
        ]

        for key, value in tags:
            body.append('    <%s>%s</%s>' % (key, value, key))

        body.append('</%s>\n' % root_tag)

        body = '\n'.join(body)

        explanation = cls._encode_explanation(unicode_explanation)
        response = Response(
            body=body,
            status_int=exc.code,
            content_type=cls.content_type,
            charset=output_encoding,
        )
        response.headers.add('Warning', explanation)

        return response

    @classmethod
    def content_location_response(cls, locations, request):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']

        if locations:
            root_tag = ini_settings.get('xml_sdb.root_tag', 'SlashDB')
            body = [
                f'<?xml version="1.0" encoding="{output_encoding}"?>',
                f'<{root_tag}>'
            ]

            locations = [f'{l}.{cls.default_extension}' for l in locations]

            for loc in locations:
                body.append(f'    <Content-Location>{loc}</Content-Location>')

            body.append(f'</{root_tag}>\n')
            body = '\n'.join(body)

            response = HTTPCreated(
                body=body,
                content_location=locations[0] if len(locations) == 1 else None,
                content_type=cls.content_type,
                charset=output_encoding,
            )
        else:
            response = HTTPCreated(
                charset=output_encoding,
                content_type=cls.content_type,
            )

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.ini_settings = request.registry.settings
        self.res_req_meta = data_proxy.res_req_meta
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        sdb_settings = request.registry.settings['sdb_settings']
        self.output_encoding = sdb_settings['output_encoding']
        self.build_escaped_url = context.get_replace_fun()
        self.xml_type = self.request.params.get('xmlType')
        self._enable_href = get_href_param(
            request.params,
            self.ini_settings.get(
                self.href_ini_param_str,
                self.href_enabled_default
            )
        )

    @property
    def nil_visible(self):
        url_param = self.request.params.get('nil_visible')
        renderer_default = self.ini_settings.get('xml_sdb.nil_visible')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter nil_visible is missing value (true or false).")
        elif url_param:
            return asbool(url_param)
        elif renderer_default:
            return asbool(renderer_default)
        else:
            return False

    def render(self):
        if self.data_proxy.transpose:
            raise HTTPBadRequest(
                explanation='Transposing for xml/xsd is not implemented'
            )
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        emtr = XMLEmitter
        is_streaming = self.is_streaming_response()

        if self.context.query_id and self.xml_type == 'adPersistXML':
            emtr = XMLRecordsetEmitter

        xml_gen = emtr(
            self.context,
            self.data_proxy,
            self.request.host_url,
            self.depth,
            self.ini_settings,
            self.nil_visible,
            self.build_escaped_url,
            is_streaming,
            enable_href=self._enable_href,
        )

        if is_streaming:
            return xml_gen

        return xml_gen._to_string()


class NoParamsXML(NoParams):
    def __init__(self, root_tag, param_set_tag):
        msg = "Unexpected data structure in incoming XML. Expected is a root element \"%s\" " \
              "with many sub-elements \"%s\" each containing tags named like parameters and with values inside." \
              % (root_tag, param_set_tag)

        super(NoParamsXML, self).__init__(msg)


class XMLParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.url_segments = context.url_segments
        self.db_name = context.db_id
        self.settings = context.settings
        self.query_id = context.query_id

        self.res_req_meta = data_proxy.res_req_meta
        self.host_url = request.host_url
        self.req_params = request.params
        self.ini_settings = request.registry.settings

    @staticmethod
    def parse_sqlpassthru_parameters(required_params, request_body, request_charset, request_get_params, ini_settings):
        root_tag = ini_settings.get('xml_sdb.sqlpassthru_params.root_tag', 'parameters')
        param_set_tag = ini_settings.get('xml_sdb.sqlpassthru_params.set_tag', 'param_set')
        required_params = set(required_params)
        parsed_data = []

        if not request_body.strip():
            if required_params:
                raise NoParams("No data to parse found in incoming XML.")
            else:
                return []

        try:
            root = ET.fromstring(request_body)

            if root.tag == root_tag:
                param_sets = root.findall('./{}'.format(param_set_tag))
                for param_set in param_sets:
                    parsed_data.append(
                        {param.tag: param.text for param in param_set}
                    )
            elif root.tag == param_set_tag:  # root omitted in request body
                parsed_data.append(
                    {param.tag: param.text for param in root}
                )
            else:
                raise NoParamsXML(root_tag, param_set_tag)
        except Exception as xml_exc:
            raise NoParams(get_message(xml_exc))

        try:
            found_param_names = set(
                chain(*(list(param_set.keys()) for param_set in parsed_data))
            )
        except ValueError:
            raise NoParamsXML(root_tag, param_set_tag)

        if required_params:
                unexpected_params = set(found_param_names) - set(required_params)
                if unexpected_params:
                    raise UnexpectedParameters(
                        'Unexpected SQL Pass-thru parameters in incoming XML.',
                        unexpected_params
                    )

                missing_params = set(required_params) - set(found_param_names)
                if missing_params:
                    raise MissingRequiredParameters(
                        "Missing SQL Pass-thru parameters in incoming XML.",
                        missing_params
                    )

                return parsed_data
        elif parsed_data:
            raise UnexpectedParameters(
                "Unexpected data found in incoming XML for SQLPass-thru that "
                "doesn't require parameters.",
                found_param_names
            )
        else:
            return []

    def parse(self, data_str):
        xsd_schema = XSDEncoder(
            self.query_id,
            self.data_proxy,
            int(self.req_params.get('depth', 0)),
            self.req_params.get('cardinality', 'unbounded'),
            self.ini_settings
        ).encode()

        xml_decoder = XMLDecoder(
            self.url_segments,
            self.res_req_meta,
            self.db_name,
            self.settings,
            self.host_url,
            self.req_params,
            xsd_schema
        )

        update_dict = xml_decoder.decode(data_str)
        return update_dict


Renderer = XMLRenderer
Parser = XMLParser
