# -*- coding: utf-8 -*-

"""Package with XSD renderer for SlashDB."""

import os
import urllib.request, urllib.parse, urllib.error

from pyramid.httpexceptions import HTTPBadRequest, HTTPNotImplemented, HTTPSeeOther
from pyramid.response import Response

from slashdb.formats import DefaultIsStreamingMixin, IRenderer, check_for_ambiguous_keys
from slashdb.formats.xsd_sdb.encoder import XSDEncoder
from slashdb.querydef_utils import make_query_url_template
from slashdb.utils import to_unicode

__version__ = '0.0.1'

EXTENSIONS = ('xsd', )
MIME_TYPES = ('text/xsd', 'application/xsd')


class XSDRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    default_extension = EXTENSIONS[0]
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = 'text/xml'
    stream_by_default_str = 'xsd_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request, **kwargs):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        root_tag = ini_settings.get('xml_sdb.root_tag', 'SlashDB')

        unicode_explanation = to_unicode(exc.explanation)
        tags = [
            ('http_code', exc.status_code),
            ('description', unicode_explanation)
        ]

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            tags.append(
                ('url_template', make_query_url_template(query_config, 'xsd'))
            )
        except (AttributeError, TypeError, KeyError):
            pass  # no additional data

        for key_value in list(kwargs.items()):
            tags.append(key_value)

        body = [
            '<?xml version="1.0" encoding="{}"?>'.format(output_encoding),
            '<%s>' % root_tag,
        ]

        for key, value in tags:
            body.append('    <%s>%s</%s>' % (key, value, key))

        body.append(
            '</%s>\n' % root_tag
        )

        body = '\n'.join(body)

        explanation = cls._encode_explanation(unicode_explanation)
        response = Response(
            body=body,
            status_int=exc.code,
            charset=output_encoding,
            content_type=cls.content_type,
        )
        response.headers.add('Warning', explanation)

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.url_segments = context.url_segments
        self.accept_header = context.accept_header
        self.db_id = context.db_id
        self.query_id = context.query_id
        self.settings = context.settings

        self.res_req_meta = data_proxy.res_req_meta
        self.is_single_resource = data_proxy.is_single_resource

        self.host = request.host
        self.host_url = request.host_url
        self.ini_settings = request.registry.settings

        self.req_params = request.params
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.cardinality = self.request.params.get('cardinality', 'unbounded')

        sdb_settings = self.ini_settings['sdb_settings']
        self.output_encoding = sdb_settings['output_encoding']
        self.applicable_query_params = (
            'depth', 'cardinality', self.settings['apikey_name'], self.settings['appid_name']
        )

    def _get_schema_url(self):
        schema_url = ''
        schema_url_qs = ''
        ext = '' if self.accept_header else '.xsd'

        if self.query_id:
            schema_url = '{}/query/{}/{}'.format(self.host_url, self.query_id, os.path.join(*self.url_segments))

        elif self.res_req_meta:
            res_req_meta = self.res_req_meta[-1]
            class_name = res_req_meta.resource_class.__name__
            query_scalars = res_req_meta.scalars

            if len(query_scalars) == 0:
                schema_path = "/db/{}/{}".format(self.db_id, class_name)
            else:
                scalar = ','.join(query_scalars)
                schema_path = "/db/{}/{}/{}".format(self.db_id, class_name, scalar)

            schema_url = self.host_url + urllib.parse.quote(schema_path)
            url_params = {}
            for param_name in self.applicable_query_params:
                value = self.request.params.get(param_name)
                if value:
                    url_params[param_name] = value

            if self.is_single_resource:
                url_params['cardinality'] = '1'

            if url_params:
                schema_url_qs = '?' + urllib.parse.urlencode(url_params)

        schema_url += ext

        return schema_url, schema_url_qs

    def render(self):
        if self.data_proxy.transpose:
            raise HTTPBadRequest(
                explanation='Transposing for xml/xsd is not implemented'
            )
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        if self.is_streaming_response():
            raise HTTPNotImplemented(explanation="Streaming XSD is not implemented.")

        xsd_maker = XSDEncoder(
            self.query_id,
            self.data_proxy,
            self.depth,
            self.cardinality,
            self.ini_settings
        )

        if self.query_id:
            return xsd_maker.encode_sqlpassthru()
        elif self.res_req_meta:
            schema_url, schema_url_qs = self._get_schema_url()
            if urllib.parse.unquote(schema_url) != urllib.parse.unquote(self.request.path_url):
                raise HTTPSeeOther(location=schema_url + schema_url_qs)

            return xsd_maker.encode()

        raise HTTPBadRequest(explanation="XSD is available only for Data Discovery requests.")

    def response_postprocessing(self, response):
        pass


Renderer = XSDRenderer
Parser = None
